# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Main application class for embedded Mechanical."""
import atexit
import os
import typing
import warnings

from ansys.mechanical.core.embedding import initializer, runtime
from ansys.mechanical.core.embedding.addins import AddinConfiguration
from ansys.mechanical.core.embedding.appdata import UniqueUserProfile
from ansys.mechanical.core.embedding.imports import global_entry_points, global_variables
from ansys.mechanical.core.embedding.poster import Poster
from ansys.mechanical.core.embedding.warnings import connect_warnings, disconnect_warnings

try:
    import ansys.tools.visualization_interface  # noqa: F401

    HAS_ANSYS_VIZ = True
    """Whether or not PyVista exists."""
except:

    HAS_ANSYS_VIZ = False


def _get_default_addin_configuration() -> AddinConfiguration:
    configuration = AddinConfiguration()
    return configuration


INSTANCES = []
"""List of instances."""


def _dispose_embedded_app(instances):  # pragma: nocover
    if len(instances) > 0:
        instance = instances[0]
        instance._dispose()


def _cleanup_private_appdata(profile: UniqueUserProfile):
    profile.cleanup()


def _start_application(configuration: AddinConfiguration, version, db_file) -> "App":
    import clr

    clr.AddReference("Ansys.Mechanical.Embedding")
    import Ansys

    if configuration.no_act_addins:
        os.environ["ANSYS_MECHANICAL_STANDALONE_NO_ACT_EXTENSIONS"] = "1"

    addin_configuration_name = configuration.addin_configuration
    # Starting with version 241 we can pass a configuration name to the constructor
    # of Application
    if int(version) >= 241:
        return Ansys.Mechanical.Embedding.Application(db_file, addin_configuration_name)
    else:
        return Ansys.Mechanical.Embedding.Application(db_file)


class GetterWrapper(object):
    """Wrapper class around an attribute of an object."""

    def __init__(self, obj, getter):
        """Create a new instance of GetterWrapper."""
        # immortal class which provides wrapped object
        self.__dict__["_immortal_object"] = obj
        # function to get the wrapped object from the immortal class
        self.__dict__["_get_wrapped_object"] = getter

    def __getattr__(self, attr):
        """Wrap getters to the wrapped object."""
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._get_wrapped_object(self._immortal_object), attr)

    def __setattr__(self, attr, value):
        """Wrap setters to the wrapped object."""
        if attr in self.__dict__:
            setattr(self, attr, value)
        setattr(self._get_wrapped_object(self._immortal_object), attr, value)


class App:
    """Mechanical embedding Application."""

    def __init__(self, db_file=None, private_appdata=False, **kwargs):
        """Construct an instance of the mechanical Application.

        db_file is an optional path to a mechanical database file (.mechdat or .mechdb)
        you may set a version number with the `version` keyword argument.

        private_appdata is an optional setting for a temporary AppData directory.
        By default, private_appdata is False. This enables you to run parallel
        instances of Mechanical.
        """
        global INSTANCES
        from ansys.mechanical.core import BUILDING_GALLERY

        if BUILDING_GALLERY:
            if len(INSTANCES) != 0:
                instance: App = INSTANCES[0]
                instance._share(self)
                if db_file != None:
                    self.open(db_file)
                return
        if len(INSTANCES) > 0:
            raise Exception("Cannot have more than one embedded mechanical instance!")
        version = kwargs.get("version")
        self._version = initializer.initialize(version)
        configuration = kwargs.get("config", _get_default_addin_configuration())

        if private_appdata:
            new_profile_name = f"PyMechanical-{os.getpid()}"
            profile = UniqueUserProfile(new_profile_name)
            profile.update_environment(os.environ)
            atexit.register(_cleanup_private_appdata, profile)

        self._app = _start_application(configuration, self._version, db_file)
        runtime.initialize(self._version)
        connect_warnings(self)
        self._poster = None

        self._disposed = False
        atexit.register(_dispose_embedded_app, INSTANCES)
        INSTANCES.append(self)
        self._updated_scopes: typing.List[typing.Dict[str, typing.Any]] = []
        self._subscribe()

    def __repr__(self):
        """Get the product info."""
        if self._version < 232:  # pragma: no cover
            return "Ansys Mechanical"
        import clr

        clr.AddReference("Ansys.Mechanical.Application")
        import Ansys

        return Ansys.Mechanical.Application.ProductInfo.ProductInfoAsString

    def __enter__(self):  # pragma: no cover
        """Enter the scope."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """Exit the scope."""
        self._dispose()

    def _dispose(self):
        if self._disposed:
            return
        self._unsubscribe()
        disconnect_warnings(self)
        self._app.Dispose()
        self._disposed = True

    def open(self, db_file):
        """Open the db file."""
        self.DataModel.Project.Open(db_file)

    def save(self, path=None):
        """Save the project."""
        if path is not None:
            self.DataModel.Project.Save(path)
        else:
            self.DataModel.Project.Save()

    def save_as(self, path):
        """Save the project as."""
        self.DataModel.Project.SaveAs(path)

    def new(self):
        """Clear to a new application."""
        self.DataModel.Project.New()

    def close(self):
        """Close the active project."""
        # Call New() to remove the lock file of the
        # current project on close.
        self.DataModel.Project.New()

    def exit(self):
        """Exit the application."""
        self._unsubscribe()
        if self.version < 241:
            self.ExtAPI.Application.Close()
        else:
            self.ExtAPI.Application.Exit()

    def execute_script(self, script: str) -> typing.Any:
        """Execute the given script with the internal IronPython engine."""
        SCRIPT_SCOPE = "pymechanical-internal"
        if not hasattr(self, "script_engine"):
            import clr

            clr.AddReference("Ansys.Mechanical.Scripting")
            import Ansys

            engine_type = Ansys.Mechanical.Scripting.ScriptEngineType.IronPython
            script_engine = Ansys.Mechanical.Scripting.EngineFactory.CreateEngine(engine_type)
            empty_scope = False
            debug_mode = False
            script_engine.CreateScope(SCRIPT_SCOPE, empty_scope, debug_mode)
            self.script_engine = script_engine
        light_mode = True
        args = None
        rets = None
        return self.script_engine.ExecuteCode(script, SCRIPT_SCOPE, light_mode, args, rets)

    def plotter(self) -> None:
        """Return ``ansys.tools.visualization_interface.Plotter`` object."""
        if not HAS_ANSYS_VIZ:
            warnings.warn(
                "Installation of viz option required! Use pip install ansys-mechanical-core[viz]"
            )
            return

        if self.version < 242:
            warnings.warn("Plotting is only supported with version 2024R2 and later!")
            return

        # TODO Check if anything loaded inside app or else show warning and return

        from ansys.mechanical.core.embedding.viz.embedding_plotter import to_plotter

        return to_plotter(self)

    def plot(self) -> None:
        """Visualize the model in 3d.

        Requires installation using the viz option. E.g.
        pip install ansys-mechanical-core[viz]
        """
        _plotter = self.plotter()
        return _plotter.show()

    @property
    def poster(self) -> Poster:
        """Returns an instance of Poster."""
        if self._poster == None:
            self._poster = Poster()
        return self._poster

    @property
    def DataModel(self):
        """Return the DataModel."""
        return GetterWrapper(self._app, lambda app: app.DataModel)

    @property
    def ExtAPI(self):
        """Return the ExtAPI object."""
        return GetterWrapper(self._app, lambda app: app.ExtAPI)

    @property
    def Tree(self):
        """Return the Tree object."""
        return GetterWrapper(self._app, lambda app: app.DataModel.Tree)

    @property
    def Model(self):
        """Return the Model object."""
        return GetterWrapper(self._app, lambda app: app.DataModel.Project.Model)

    @property
    def Graphics(self):
        """Return the Graphics object."""
        return GetterWrapper(self._app, lambda app: app.ExtAPI.Graphics)

    @property
    def readonly(self):
        """Return whether the Mechanical object is read-only."""
        import Ansys

        return Ansys.ACT.Mechanical.MechanicalAPI.Instance.ReadOnlyMode

    @property
    def version(self):
        """Returns the version of the app."""
        return self._version

    def _share(self, other) -> None:
        """Shares the state of self with other.

        Other is another instance of App.
        This is used when the BUILDING_GALLERY flag is on.
        In that mode, multiple instance of App are used, but
        they all point to the same underlying application
        object. Because of that, special care needs to be
        taken to properly share the state. Other will be
        a "weak reference", which doesn't own anything.
        """
        # the other app is not expecting to have a project
        # already loaded
        self.new()

        # set up the type hint (typing.Self is python3.11+)
        other: App = other

        # copy `self` state to other.
        other._app = self._app
        other._version = self._version
        other._poster = self._poster
        other._updated_scopes = self._updated_scopes

        # all events will be handled by the original App instance
        other._subscribed = False

        # finally, set the other disposed flag to be true
        # so that the shutdown sequence isn't duplicated
        other._disposed = True

    def _subscribe(self):
        try:
            # This will throw an error when using pythonnet because
            # EventSource isn't defined on the IApplication interface
            self.ExtAPI.Application.EventSource.OnWorkbenchReady += self._on_workbench_ready
            self._subscribed = True
        except:
            self._subscribed = False

    def _unsubscribe(self):
        if not self._subscribed:
            return
        self._subscribed = False
        self.ExtAPI.Application.EventSource.OnWorkbenchReady -= self._on_workbench_ready

    def _on_workbench_ready(self, sender, args) -> None:
        self._update_all_globals()

    def update_globals(
        self, globals_dict: typing.Dict[str, typing.Any], enums: bool = True
    ) -> None:
        """Use to update globals variables.

        When scripting inside Mechanical, the Mechanical UI will automatically
        set global variables in python. PyMechanical can not do that automatically,
        but this method can be used.
        `app.update_globals(globals())`

        By default, all enums will be imported too. To avoid including enums, set
        the `enums` argument to False.
        """
        self._updated_scopes.append(globals_dict)
        globals_dict.update(global_variables(self, enums))

    def _update_all_globals(self) -> None:
        for scope in self._updated_scopes:
            scope.update(global_entry_points(self))

    def _print_tree(self, node, max_lines, lines_count, indentation):
        """Recursively print till provided maximum lines limit.

        Each object in the tree is expected to have the following attributes:
         - Name: The name of the object.
         - Suppressed : Print as suppressed, if object is suppressed.
         - Children: Checks if object have children.
           Each child node is expected to have the all these attributes.

        Parameters
        ----------
        lines_count: int, optional
            The current count of lines printed. Default is 0.
        indentation: str, optional
            The indentation string used for printing the tree structure. Default is "".
        """
        if lines_count >= max_lines and max_lines != -1:
            print(f"... truncating after {max_lines} lines")
            return lines_count

        if not hasattr(node, "Name"):
            raise AttributeError("Object must have a 'Name' attribute")

        node_name = node.Name
        if hasattr(node, "Suppressed") and node.Suppressed is True:
            node_name += " (Suppressed)"
        print(f"{indentation}├── {node_name}")
        lines_count += 1

        if lines_count >= max_lines and max_lines != -1:
            print(f"... truncating after {max_lines} lines")
            return lines_count

        if hasattr(node, "Children") and node.Children is not None and node.Children.Count > 0:
            for child in node.Children:
                lines_count = self._print_tree(child, max_lines, lines_count, indentation + "|  ")
                if lines_count >= max_lines and max_lines != -1:
                    break

        return lines_count

    def print_tree(self, node=None, max_lines=80, lines_count=0, indentation=""):
        """
        Print the hierarchical tree representation of the Mechanical project structure.

        Parameters
        ----------
        node: DataModel object, optional
            The starting object of the tree.
        max_lines: int, optional
            The maximum number of lines to print. Default is 80. If set to -1, no limit is applied.

        Raises
        ------
        AttributeError
            If the node does not have the required attributes.

        Examples
        --------
        >>> import ansys.mechanical.core as mech
        >>> app = mech.App()
        >>> app.update_globals(globals())
        >>> app.print_tree()
        ... ├── Project
        ... |  ├── Model
        ... |  |  ├── Geometry Imports
        ... |  |  ├── Geometry
        ... |  |  ├── Materials
        ... |  |  ├── Coordinate Systems
        ... |  |  |  ├── Global Coordinate System
        ... |  |  ├── Remote Points
        ... |  |  ├── Mesh

        >>> app.print_tree(Model, 3)
        ... ├── Model
        ... |  ├── Geometry Imports
        ... |  ├── Geometry
        ... ... truncating after 3 lines

        >>> app.print_tree(max_lines=2)
        ... ├── Project
        ... |  ├── Model
        ... ... truncating after 2 lines
        """
        if node is None:
            node = self.DataModel.Project

        self._print_tree(node, max_lines, lines_count, indentation)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Initialize the package level imports."""
import logging
import os

from ansys.tools.path import find_mechanical
import appdirs

USER_DATA_PATH = appdirs.user_data_dir(appname="ansys_mechanical_core", appauthor="Ansys")
"""User data directory."""

if not os.path.exists(USER_DATA_PATH):
    os.makedirs(USER_DATA_PATH)

EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
"""Examples path."""

if not os.path.exists(EXAMPLES_PATH):
    os.makedirs(EXAMPLES_PATH)

from ansys.mechanical.core.logging import Logger

LOG = Logger(level=logging.ERROR, to_file=False, to_stdout=True)
"""Create logger for package level use."""

LOG.debug("Loaded logging module as LOG")

from ansys.mechanical.core._version import __version__
from ansys.mechanical.core.mechanical import (
    change_default_mechanical_path,
    close_all_local_instances,
    get_mechanical_path,
    launch_mechanical,
)

# import few classes / functions
from ansys.mechanical.core.mechanical import Mechanical as Mechanical

try:
    from ansys.mechanical.core.embedding import App, global_variables

    HAS_EMBEDDING = True
    """Whether or not Mechanical embedding is being used."""
except:
    HAS_EMBEDDING = False

LOCAL_PORTS = []
"""Manage the package level ports."""

from ansys.mechanical.core.pool import LocalMechanicalPool

BUILDING_GALLERY = False
"""Whether or not to build gallery examples."""



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Class for running Mechanical on a background thread."""

import atexit
import threading
import time
import typing

import ansys.mechanical.core as mech
from ansys.mechanical.core.embedding.poster import Poster
import ansys.mechanical.core.embedding.utils as utils


def _exit(background_app: "BackgroundApp"):
    """Stop the thread serving the Background App."""
    background_app.stop()
    atexit.unregister(_exit)


class BackgroundApp:
    """Background App."""

    __app: mech.App = None
    __app_thread: threading.Thread = None
    __stopped: bool = False
    __stop_signaled: bool = False
    __poster: Poster = None

    def __init__(self, **kwargs):
        """Construct an instance of BackgroundApp."""
        if BackgroundApp.__app_thread == None:
            BackgroundApp.__app_thread = threading.Thread(
                target=self._start_app, kwargs=kwargs, daemon=True
            )
            BackgroundApp.__app_thread.start()

            while BackgroundApp.__poster is None:
                time.sleep(0.05)
                continue
        else:
            assert (
                not BackgroundApp.__stopped
            ), "Cannot initialize a BackgroundApp once it has been stopped!"

            def new():
                BackgroundApp.__app.new()

            self.post(new)

        atexit.register(_exit, self)

    @property
    def app(self) -> mech.App:
        """Get the App instance of the background thread.

        It is not meant to be used aside from passing to methods using `post`.
        """
        return BackgroundApp.__app

    def post(self, callable: typing.Callable):
        """Post callable method to the background app thread."""
        assert not BackgroundApp.__stopped, "Cannot use background app after stopping it."
        return BackgroundApp.__poster.post(callable)

    def stop(self) -> None:
        """Stop the background app thread."""
        if BackgroundApp.__stopped:
            return
        BackgroundApp.__stop_signaled = True
        while True:
            time.sleep(0.05)
            if BackgroundApp.__stopped:
                break

    def _start_app(self, **kwargs) -> None:
        BackgroundApp.__app = mech.App(**kwargs)
        BackgroundApp.__poster = BackgroundApp.__app.poster
        while True:
            if BackgroundApp.__stop_signaled:
                break
            try:
                utils.sleep(40)
            except:
                pass
        BackgroundApp.__stopped = True



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Common plotting utilities."""
import typing

import numpy as np


def bgr_to_rgb_tuple(bgr_int: int) -> typing.Tuple[int, int, int]:
    """Convert bgr integer to rgb tuple."""
    r = bgr_int & 255
    g = (bgr_int >> 8) & 255
    b = (bgr_int >> 16) & 255
    return r, g, b


def _reshape_3cols(arr: np.array, name: str = "array"):
    """Reshapes the given array into 3 columns.

    Precondition - the array's length must be divisible by 3.
    """
    err = f"{name} must be of the form (x0,y0,z0,x1,y1,z1,...,xn,yn,zn).\
        Given {name} are not divisible by 3!"
    assert arr.size % 3 == 0, err
    numrows = int(arr.size / 3)
    numcols = 3
    arr = np.reshape(arr, (numrows, numcols))
    return arr


def get_nodes_and_coords(tri_tessellation: "Ansys.Mechanical.Scenegraph.TriTessellationNode"):
    """Extract the nodes and coordinates from the TriTessellationNode.

    The TriTessellationNode contains "Coordinates" and "Indices"
    that are flat arrays. This function converts them to numpy arrays
    """
    np_coordinates = _reshape_3cols(
        np.array(tri_tessellation.Coordinates, dtype=np.double), "coordinates"
    )
    np_indices = _reshape_3cols(np.array(tri_tessellation.Indices, dtype=np.int32), "indices")
    return np_coordinates, np_indices



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Convenience CLI to run mechanical."""

import asyncio
from asyncio.subprocess import PIPE
import os
import sys
import typing
import warnings

import ansys.tools.path as atp
import click

from ansys.mechanical.core.embedding.appdata import UniqueUserProfile
from ansys.mechanical.core.feature_flags import get_command_line_arguments, get_feature_flag_names

DRY_RUN = False
"""Dry run constant."""

# TODO - add logging options (reuse env var based logging initialization)
# TODO - add timeout


async def _read_and_display(cmd, env, do_display: bool):
    """Read command's stdout and stderr and display them as they are processed."""
    # start process
    process = await asyncio.create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE, env=env)
    # read child's stdout/stderr concurrently
    stdout, stderr = [], []  # stderr, stdout buffers
    tasks = {
        asyncio.Task(process.stdout.readline()): (stdout, process.stdout, sys.stdout.buffer),
        asyncio.Task(process.stderr.readline()): (stderr, process.stderr, sys.stderr.buffer),
    }
    while tasks:
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        assert done
        for future in done:
            buf, stream, display = tasks.pop(future)
            line = future.result()
            if line:  # not EOF
                buf.append(line)  # save for later
                if do_display:
                    display.write(line)  # display in terminal
                # schedule to read the next line
                tasks[asyncio.Task(stream.readline())] = buf, stream, display

    # wait for the process to exit
    rc = await process.wait()
    return rc, b"".join(stdout), b"".join(stderr)


def _run(args, env, check=False, display=False):
    if os.name == "nt":
        loop = asyncio.ProactorEventLoop()  # for subprocess' pipes on Windows
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.get_event_loop()
    try:
        rc, *output = loop.run_until_complete(_read_and_display(args, env, display))
        if rc and check:
            sys.exit("child failed with '{}' exit code".format(rc))
    finally:
        if os.name == "nt":
            loop.close()
    return output


def _cli_impl(
    project_file: str = None,
    port: int = 0,
    debug: bool = False,
    input_script: str = None,
    script_args: str = None,
    exe: str = None,
    version: int = None,
    graphical: bool = False,
    show_welcome_screen: bool = False,
    private_appdata: bool = False,
    exit: bool = False,
    features: str = None,
):
    if project_file and input_script:
        raise Exception("Cannot open a project file *and* run a script.")

    if (not graphical) and project_file:
        raise Exception("Cannot open a project file in batch mode.")

    if port:
        if project_file:
            raise Exception("Cannot open in server mode with a project file.")
        if input_script:
            raise Exception("Cannot open in server mode with an input script.")

    if not input_script and script_args:
        raise Exception("Cannot add script arguments without an input script.")

    if script_args:
        if '"' in script_args:
            raise Exception(
                "Cannot have double quotes around individual arguments in the --script-args string."
            )

    # If the input_script and port are missing in batch mode, raise an exception
    if (not graphical) and (input_script is None) and (not port):
        raise Exception("An input script, -i, or port, --port, are required in batch mode.")

    args = [exe, "-DSApplet"]
    if (not graphical) or (not show_welcome_screen):
        args.append("-AppModeMech")

    if version < 232:
        args.append("-nosplash")
        args.append("-notabctrl")

    if not graphical:
        args.append("-b")

    env: typing.Dict[str, str] = os.environ.copy()
    if debug:
        env["WBDEBUG_STOP"] = "1"

    if port:
        args.append("-grpc")
        args.append(str(port))

    if project_file:
        args.append("-file")
        args.append(project_file)

    if input_script:
        args.append("-script")
        args.append(input_script)

    if script_args:
        args.append("-ScriptArgs")
        args.append(f'"{script_args}"')

    if (not graphical) and input_script:
        exit = True
        if version < 241:
            warnings.warn(
                "Please ensure ExtAPI.Application.Close() is at the end of your script. "
                "Without this command, Batch mode will not terminate.",
                stacklevel=2,
            )

    if exit and input_script and version >= 241:
        args.append("-x")

    profile: UniqueUserProfile = None
    if private_appdata:
        new_profile_name = f"Mechanical-{os.getpid()}"
        profile = UniqueUserProfile(new_profile_name, DRY_RUN)
        profile.update_environment(env)

    if not DRY_RUN:
        version_name = atp.SUPPORTED_ANSYS_VERSIONS[version]
        if graphical:
            mode = "Graphical"
        else:
            mode = "Batch"
        print(f"Starting Ansys Mechanical version {version_name} in {mode} mode...")
        if port:
            # TODO - Mechanical doesn't write anything to the stdout in grpc mode
            #        when logging is off.. Ideally we let Mechanical write it, so
            #        the user only sees the message when the server is ready.
            print(f"Serving on port {port}")

    if features is not None:
        args.extend(get_command_line_arguments(features.split(";")))

    if DRY_RUN:
        return args, env
    else:
        _run(args, env, False, True)

    if private_appdata:
        profile.cleanup()


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-p",
    "--project-file",
    default=None,
    help="Opens Mechanical project file (.mechdb). Cannot be mixed with -i",
)
@click.option(
    "--private-appdata",
    default=None,
    is_flag=True,
    help="Make the appdata folder private.\
 This enables you to run parallel instances of Mechanical.",
)
@click.option(
    "--port",
    type=int,
    help="Start mechanical in server mode with the given port number",
)
@click.option(
    "--features",
    type=str,
    default=None,
    help=f"Beta feature flags to set, as a semicolon delimited list.\
 Options: {get_feature_flag_names()}",
)
@click.option(
    "-i",
    "--input-script",
    default=None,
    help="Name of the input Python script. Cannot be mixed with -p",
)
@click.option(
    "--script-args",
    default=None,
    help='Arguments to pass into the --input-script, -i. \
Write the arguments as a string, with each argument \
separated by a comma. For example, --script-args "arg1,arg2" \
This can only be used with the --input-script argument.',
)
@click.option(
    "--exit",
    is_flag=True,
    default=None,
    help="Exit the application after running an input script. \
You can only use this command with --input-script argument (-i). \
The command defaults to true you are not running the application in graphical mode. \
The ``exit`` command is only supported in version 2024 R1 or later.",
)
@click.option(
    "-s",
    "--show-welcome-screen",
    is_flag=True,
    default=False,
    help="Show the welcome screen, where you can select the file to open.\
 Only affects graphical mode",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Show a debug dialog right when the process starts.",
)
@click.option(
    "-r",
    "--revision",
    default=None,
    type=int,
    help='Ansys Revision number, e.g. "242" or "241". If none is specified\
, uses the default from ansys-tools-path',
)
@click.option(
    "-g",
    "--graphical",
    is_flag=True,
    default=False,
    help="Graphical mode",
)
def cli(
    project_file: str,
    port: int,
    debug: bool,
    input_script: str,
    script_args: str,
    revision: int,
    graphical: bool,
    show_welcome_screen: bool,
    private_appdata: bool,
    exit: bool,
    features: str,
):
    """CLI tool to run mechanical.

    USAGE:

    The following example demonstrates the main use of this tool:

        $ ansys-mechanical -r 242 -g

        Starting Ansys Mechanical version 2024R2 in graphical mode...
    """
    exe = atp.get_mechanical_path(allow_input=False, version=revision)
    version = atp.version_from_path("mechanical", exe)

    return _cli_impl(
        project_file,
        port,
        debug,
        input_script,
        script_args,
        exe,
        version,
        graphical,
        show_welcome_screen,
        private_appdata,
        exit,
        features,
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Use the Poster class to post functions to Mechanical's main thread."""

import typing


class Poster:
    """Class which can post a python callable function to Mechanical's main thread."""

    def __init__(self):
        """Create a new instance of Poster."""
        import clr

        clr.AddReference("Ans.Common.WB1ManagedUtils")
        import Ans

        self._poster = Ans.Common.WB1ManagedUtils.TaskPoster

    def post(self, callable: typing.Callable):
        """Post the callable to Mechanical's main thread.

        The main thread needs to be receiving posted messages
        in order for this to work from a background thread. Use
        the `sleep` routine from the `utils` module to make
        Mechanical available to receive messages.

        Returns the result of `callable` if any.
        """
        import System

        func = System.Func[System.Object](callable)
        return self._poster.Get[System.Object](func)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Initializer for Mechanical embedding. Sets up paths and resolvers."""
from importlib.metadata import distribution
import os
from pathlib import Path
import platform
import sys
import warnings

import ansys.tools.path as atp

from ansys.mechanical.core.embedding.loader import load_clr
from ansys.mechanical.core.embedding.resolver import resolve

INITIALIZED_VERSION = None
"""Constant for the initialized version."""

SUPPORTED_MECHANICAL_EMBEDDING_VERSIONS_WINDOWS = {242: "2024R2", 241: "2024R1", 232: "2023R2"}
"""Supported Mechanical embedding versions on Windows."""


def __add_sys_path(version: int) -> str:
    install_path = Path(os.environ[f"AWP_ROOT{version}"])
    platform_string = "winx64" if os.name == "nt" else "linx64"
    bin_path = install_path / "aisol" / "bin" / platform_string
    sys.path.append(str(bin_path.resolve()))


def __workaround_material_server(version: int) -> None:
    """Workaround material server bug in 2024 R1.

    A REST server is used as a backend for the material model GUI.
    In 2024 R1, this GUI is used even in batch mode. The server
    starts by default on a background thread, which may lead to
    a race condition on shutdown.
    """
    if version in [241]:
        os.environ["ENGRDATA_SERVER_SERIAL"] = "1"


def _get_default_linux_version() -> int:
    """Try to get the active linux version from the environment.

    On linux, embedding is only possible by setting environment variables before starting python.
    The version will then be fixed  to a specific version, based on those env vars.
    The documented way to set those variables is to run python using the ``mechanical-env`` script,
    which can be used after installing the ``ansys-mechanical-env`` package with this command:
    ``pip install ansys-mechanical-env``. The script takes user input of a version. If the user
    does not provide a version, the ``find_mechanical()`` function from the ``ansys-tools-path``
    package is used to find a version of Mechanical.
    """
    supported_versions = [232, 241, 242]
    awp_roots = {ver: os.environ.get(f"AWP_ROOT{ver}", "") for ver in supported_versions}
    installed_versions = {
        ver: path for ver, path in awp_roots.items() if path and os.path.isdir(path)
    }
    assert len(installed_versions) == 1, "multiple AWP_ROOT environment variables found!"
    return next(iter(installed_versions))


def _get_default_version() -> int:
    if os.name == "posix":
        return _get_default_linux_version()

    if os.name != "nt":  # pragma: no cover
        raise Exception("Unexpected platform!")

    _, version = atp.find_mechanical(
        supported_versions=SUPPORTED_MECHANICAL_EMBEDDING_VERSIONS_WINDOWS
    )

    # version is of the form 23.2
    int_version = int(str(version).replace(".", ""))
    return int_version


def __check_python_interpreter_architecture():
    """Embedding support only 64 bit architecture."""
    if platform.architecture()[0] != "64bit":
        raise Exception("Mechanical Embedding requires a 64-bit Python environment.")


def __check_for_mechanical_env():
    """Embedding in linux platform must use mechanical-env."""
    if platform.system() == "Linux" and os.environ.get("PYMECHANICAL_EMBEDDING") != "TRUE":
        raise Exception(
            "On linux, embedding an instance of the Mechanical process using"
            "the App class requires running python inside of a Mechanical environment."
            "Use the `mechanical-env` script to do this. For more information, refer to:"
            "https://mechanical.docs.pyansys.com/version/stable/"
            "getting_started/running_mechanical.html#embed-a-mechanical-instance"
        )


def initialize(version: int = None):
    """Initialize Mechanical embedding."""
    __check_python_interpreter_architecture()  # blocks 32 bit python
    __check_for_mechanical_env()  # checks for mechanical-env in linux embedding

    global INITIALIZED_VERSION
    if INITIALIZED_VERSION != None:
        assert INITIALIZED_VERSION == version
        return

    if version == None:
        version = _get_default_version()

    INITIALIZED_VERSION = version

    __workaround_material_server(version)

    # need to add system path in order to import the assembly with the resolver
    __add_sys_path(version)

    # Check if 'pythonnet' is installed... and if so, throw warning
    try:
        distribution("pythonnet")
        warnings.warn(
            "The pythonnet package was found in your environment "
            "which interferes with the ansys-pythonnet package. "
            "Some APIs may not work due to pythonnet being installed.\n\n"
            "For using PyMechanical, we recommend you do the following:\n"
            "1. Uninstall your existing pythonnet package: pip uninstall pythonnet\n"
            "2. Install the ansys-pythonnet package: pip install --upgrade "
            "--force-reinstall ansys-pythonnet\n",
            stacklevel=2,
        )
    except ModuleNotFoundError:
        pass

    # load the CLR with mono that is shipped with the unified ansys installer
    load_clr(os.environ[f"AWP_ROOT{version}"])

    # attach the resolver
    resolve(version)

    return version



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PyVista plotter."""

import clr

clr.AddReference("Ansys.Mechanical.DataModel")
clr.AddReference("Ansys.ACT.Interfaces")

import Ansys  # isort: skip

from ansys.tools.visualization_interface import Plotter
import numpy as np
import pyvista as pv

from .utils import bgr_to_rgb_tuple, get_nodes_and_coords


def _transform_to_pyvista(transform: "Ansys.ACT.Math.Matrix4D"):
    """Convert the Transformation matrix to a numpy array."""
    np_transform = np.array([transform[i] for i in range(16)]).reshape(4, 4)

    # The mechanical scenegraph transform node is the transpose of the pyvista transform matrix
    np_transform = np_transform.transpose()
    return np_transform


def _get_tri_nodes_and_coords(tri_tessellation: "Ansys.Mechanical.Scenegraph.TriTessellationNode"):
    """Get the nodes and indices of the TriTessellationNode.

    pyvista format expects a number of vertices per facet which is always 3
    from this kind of node.
    """
    np_coordinates, np_indices = get_nodes_and_coords(tri_tessellation)
    np_indices = np.insert(np_indices, 0, 3, axis=1)
    return np_coordinates, np_indices


def _get_nodes_and_coords(node: "Ansys.Mechanical.Scenegraph.Node"):
    """Get the nodes and indices of the Scenegraph node.

    Currently only supported for tri tessellation nodes
    """
    if isinstance(node, Ansys.Mechanical.Scenegraph.TriTessellationNode):
        return _get_tri_nodes_and_coords(node)

    # TODO - support line tessellation node. See issue #809
    # if isinstance(node, Ansys.Mechanical.Scenegraph.LineTessellationNode):
    return None, None


def to_plotter(app: "ansys.mechanical.core.embedding.App"):
    """Convert the app's geometry to an ``ansys.tools.visualization_interface.Plotter`` instance."""
    plotter = Plotter()
    for body in app.DataModel.GetObjectsByType(
        Ansys.Mechanical.DataModel.Enums.DataModelObjectCategory.Body
    ):
        scenegraph_node = Ansys.ACT.Mechanical.Tools.ScenegraphHelpers.GetScenegraph(body)
        np_coordinates, np_indices = _get_nodes_and_coords(scenegraph_node.Child)
        if np_coordinates is None or np_indices is None:
            continue
        pv_transform = _transform_to_pyvista(scenegraph_node.Transform)
        polydata = pv.PolyData(np_coordinates, np_indices).transform(pv_transform)
        color = pv.Color(bgr_to_rgb_tuple(body.Color))
        plotter.plot(polydata, color=color, smooth_shading=True)
    return plotter



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Launch Mechanical in batch or UI mode."""
import errno
import os
import subprocess

from ansys.mechanical.core import LOG
from ansys.mechanical.core.misc import is_windows


class MechanicalLauncher:
    """Launches Mechanical in batch or UI mode."""

    def __init__(
        self, batch, port, exe_path, additional_args=None, additional_envs=None, verbose=False
    ):
        """Initialize the Mechanical launcher.

        Parameters
        ----------
        batch : bool, optional
            Whether to launch Mechanical in batch mode. The default is ``True``. When ``False``,
            Mechanical is launched in UI mode.
        port : int, optional
            Port to connect to the Mechanical gRPC server. Default - find a free port
        exe_path : str
            Path to where the executable file for Mechanical is located.
        additional_args : list, optional
            List of additional arguments to pass. The default is ``None``.
        additional_envs : dict, optional
            List of additional environment variables to pass. The default is ``None``.
        verbose : bool, optional
            Whether to print all output when launching and running Mechanical. The
            default is ``False``. Only set this parameter to ``True`` if you are
            debugging the startup of Mechanical.
        """
        self.batch = batch
        self.port = port
        self.exe_path = exe_path
        self.additional_args = additional_args
        self.additional_envs = additional_envs
        self.verbose = verbose
        self.__ui_arg_list = ["-DSApplet", "-nosplash", "-notabctrl"]
        self.__batch_arg_list = ["-DSApplet", "-b"]

        app_mode_mech_exits = MechanicalLauncher._mode_exists(additional_args, "-AppModeMech")
        app_mode_mesh_exits = MechanicalLauncher._mode_exists(additional_args, "-AppModeMesh")
        app_mode_rest_exits = MechanicalLauncher._mode_exists(additional_args, "-AppModeRest")

        # if we don't have -AppModeMesh or -AppModeRest or -AppModeMech in the additional args
        # then we want to start with -AppModeMech
        if not (app_mode_mesh_exits or app_mode_rest_exits or app_mode_mech_exits):
            self.__ui_arg_list.append("-AppModeMech")
            self.__batch_arg_list.append("-AppModeMech")

    @staticmethod
    def _mode_exists(additional_args, mode):
        if additional_args is not None and isinstance(additional_args, list):
            if mode.upper() in (mode_temp.upper() for mode_temp in additional_args):
                return True

        return False

    def launch(self):
        """Launch Mechanical with the gRPC server."""
        exe_path = self.__get_exe_path()
        MechanicalLauncher.verify_path_exists(exe_path)

        env_variables = self.__get_env_variables()
        args_list = self.__get_commandline_args()

        shell_value = False

        if is_windows():
            shell_value = True

        LOG.info(f"Starting the process using {args_list}.")
        if self.verbose:
            command = " ".join(args_list)
            print(f"Running {command}.")

        process = subprocess.Popen(args_list, shell=shell_value, env=env_variables)
        LOG.info(f"Started the process:{process} using {args_list}.")

    @staticmethod
    def verify_path_exists(exe_path):
        """Throw an exception if the given exe_path does not exist.

        Parameters
        ----------
        exe_path : str
            Path to verify.
        """
        if not os.path.exists(exe_path):
            LOG.info(f"Startup file:{exe_path} doesn't exist.")
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), exe_path)

    def __get_env_variables(self):
        """Get the dictionary of environment variables used while launching Mechanical.

        Returns
        -------
        dict
            Dictionary of environment variables.
        """
        default_env = dict(WB1_STANDALONE="1")
        if self.additional_envs is not None and isinstance(self.additional_envs, dict):
            LOG.info(f"Using these additional env: {self.additional_envs}.")
            ansyswbu_env = {**os.environ, **default_env, **self.additional_envs}
        else:
            ansyswbu_env = {**os.environ, **default_env}

        return ansyswbu_env

    def __get_commandline_args(self):
        """Get the list of command-line arguments used while launching Mechanical.

        Returns
        -------
        list
            List of command-line arguments.
        """
        args_list = []

        exe_path = self.__get_exe_path()
        args_list.append(exe_path)

        if self.batch:
            args_list.extend(self.__batch_arg_list)
        else:  # pragma: no cover
            args_list.extend(self.__ui_arg_list)

        args_list.append("-grpc")
        args_list.append(str(self.port))

        if self.additional_args is not None and isinstance(self.additional_args, list):
            LOG.info(f"Using these additional args: {self.additional_args}.")
            args_list.extend(self.additional_args)

        return args_list

    def __get_exe_path(self):
        """Get the path to the executable file used to launch Mechanical.

        Returns
        -------
        str
            Path to the executable file for launching Mechanical.
        """
        exe_path = None
        if self.exe_path is not None and isinstance(self.exe_path, str):
            exe_path = self.exe_path

        return exe_path



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Converter to OpenUSD."""

import typing

import clr
from pxr import Gf, Usd, UsdGeom

clr.AddReference("Ansys.Mechanical.DataModel")
clr.AddReference("Ansys.ACT.Interfaces")

import Ansys  # isort: skip

from .utils import bgr_to_rgb_tuple, get_nodes_and_coords


def _transform_to_rotation_translation(
    transform: "Ansys.ACT.Math.Matrix4D",
) -> typing.Tuple[Gf.Quatf, Gf.Vec3f]:
    """Convert the Transformation matrix to a single-precision quaternion."""
    transforms = [transform[i] for i in range(16)]
    m = Gf.Matrix4d()
    m.SetRow(0, transforms[0:4])
    m.SetRow(1, transforms[4:8])
    m.SetRow(2, transforms[8:12])
    m.SetRow(3, transforms[12:16])

    # Get quaternion from transformation matrix
    quatd: Gf.Quatd = m.ExtractRotationQuat()

    # Get translation from transformation matrix
    transd: Gf.Vec3d = m.ExtractTranslation()

    # Return as single precision
    return Gf.Quatf(quatd), Gf.Vec3f(transd)


def _convert_tri_tessellation_node(
    node: "Ansys.Mechanical.Scenegraph.TriTessellationNode",
    stage: Usd.Stage,
    path: str,
    rgb: typing.Tuple[int, int, int],
) -> Usd.Prim:
    """Convert a mechanical TriTessellationNode node into a Usd Mesh prim."""
    mesh_prim = UsdGeom.Mesh.Define(stage, path)
    np_coordinates, np_indices = get_nodes_and_coords(node)
    mesh_prim.CreatePointsAttr(np_coordinates)
    mesh_prim.CreateFaceVertexCountsAttr([3] * len(np_indices))
    mesh_prim.CreateFaceVertexIndicesAttr(np_indices)
    r, g, b = rgb
    mesh_prim.GetDisplayColorAttr().Set([(r / 255, g / 255, b / 255)])
    return mesh_prim


def _create_prim_with_transform(
    stage: Usd.Stage, path: str, node: "Ansys.Mechanical.Scenegraph.TransformNode"
) -> Usd.Prim:
    """Create an empty Usd Xform prim based on a mechanical transform node."""
    prim = UsdGeom.Xform.Define(stage, path)
    rotation, translation = _transform_to_rotation_translation(node.Transform)
    prim.AddOrientOp().Set(rotation)
    prim.AddTranslateOp().Set(translation)
    return prim


def _convert_transform_node(
    node: "Ansys.Mechanical.Scenegraph.TransformNode",
    stage: Usd.Stage,
    path: str,
    rgb: typing.Tuple[int, int, int],
) -> None:
    """Add a Usd prim to the stage based on the given mechanical transform node.

    Currently only supports transforms that contain a single tri tessellation node.
    """
    child_node = node.Child
    if isinstance(child_node, Ansys.Mechanical.Scenegraph.TriTessellationNode):
        prim = _create_prim_with_transform(stage, path, node)
        child_path = prim.GetPath().AppendPath("TriTessellation")
        _convert_tri_tessellation_node(child_node, stage, child_path, rgb)


def to_usd_stage(app: "ansys.mechanical.core.embedding.App", name: str) -> None:
    """Convert mechanical scene to usd stage."""
    stage = Usd.Stage.CreateNew(name)

    root_prim = UsdGeom.Xform.Define(stage, "/root")

    category = Ansys.Mechanical.DataModel.Enums.DataModelObjectCategory.Body
    bodies = app.DataModel.GetObjectsByType(category)
    for body in bodies:
        scenegraph_node = Ansys.ACT.Mechanical.Tools.ScenegraphHelpers.GetScenegraph(body)
        body_path = root_prim.GetPath().AppendPath(f"body{body.ObjectId}")
        _convert_transform_node(scenegraph_node, stage, body_path, bgr_to_rgb_tuple(body.Color))

    return stage


def to_usd_file(app, path: str) -> None:
    """Export mechanical scene to usd file."""
    stage = to_usd_stage(app, path)
    stage.GetRootLayer().Save()



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Logging module.

This module supplies the general framework for logging in PyMechanical. This module is
built upon the `logging <https://docs.python.org/3/library/logging.html>`_ package.
The intent is not for this module to replace the ``logging`` package but rather to provide
a way for the ``logging`` package and PyMechancial to interact.

The loggers used in the module include the name of the instance, which
is intended to be unique.  This name is printed in all the active
outputs and is used to track the different Mechanical instances.


Usage
-----

Global logger
~~~~~~~~~~~~~
There is a global logger named ``pymechanical_global``, which is created at
``ansys.mechanical.core.__init__``.  If you want to use this global logger,
you must call it at the top of your module:

.. code:: python

   from ansys.mechanical.core import LOG

You can rename this logger to avoid conflicts with other loggers (if any):

.. code:: python

   from ansys.mechanical.core import LOG as logger


The default logging level of ``LOG`` is ``ERROR``. To change this and output
lower-level messages, you can use this code:

.. code:: python

   LOG.logger.setLevel("DEBUG")
   LOG.file_handler.setLevel("DEBUG")  # If present.
   LOG.stdout_handler.setLevel("DEBUG")  # If present.


Alternatively, you can use this code:

.. code:: python

   LOG.setLevel("DEBUG")

This alternative code ensures that all the handlers are set to the
input log level.

By default, this logger does not log to a file. If you want,
you can add a file handler:

.. code:: python

   import os

   file_path = os.path.join(os.getcwd(), "pymechanical.log")
   LOG.log_to_file(file_path)

The preceding code sets the logger to also be redirected to this file. If you
want to change the characteristics of this global logger from the beginning
of the execution, you must edit the file ``__init__`` in the
``ansys.mechanical.core`` directory.

To log using this logger, call the desired method as a normal logger:

.. code:: pycon

    >>> import logging
    >>> from ansys.mechanical.core.logging import Logger
    >>> LOG = Logger(level=logging.DEBUG, to_file=False, to_stdout=True)
    >>> LOG.debug("This is LOG debug message.")

    DEBUG -  -  <ipython-input-24-80df150fe31f> - <module> - This is the LOG debug message.


Instance Logger
~~~~~~~~~~~~~~~
Every time an instance of the :class:`Mechanical <ansys.mechanical.core.mechanical.Mechanical>`
ckass is created, a logger is created and stored here:

* ``LOG._instances``. This field is a ``dict`` where the key is the name of the
  created logger.

These logger instances inherit the ``pymechanical_global`` output handlers and
logging level unless otherwise specified. The way this logger works is very
similar to the global logger. You can add a file handler if you want using the
:func:`log_to_file() <PyMechanicalCustomAdapter.log_to_file>` method or change
the log level using the :func:`logger.Logging.setLevel` method.

You can use this logger like this:

.. code:: pycon

    >>> from ansys.mechanical.core import launch_mechanical
    >>> mechanical = launch_mechanical()
    >>> mechanical.log.info("This is a useful message")

    INFO - GRPC_127.0.0.1:50056 -  <ipython-input-19-f09bb2d8785c> - <module> -
    This is a useful message

Other loggers
~~~~~~~~~~~~~
You can create your own loggers using the Python ``logging`` package as
you would do in any other script. There are no conflicts between these loggers.

"""

from copy import copy
from datetime import datetime
import logging
import sys
import weakref

# Default configuration
LOG_LEVEL = logging.DEBUG
"""Default log level configuration."""
FILE_NAME = "pymechanical.log"
"""Default file name."""

# For convenience
DEBUG = logging.DEBUG
"""Constant for logging.DEBUG."""
INFO = logging.INFO
"""Constant for logging.INFO."""
WARN = logging.WARN
"""Constant for logging.WARN."""
ERROR = logging.ERROR
"""Constant for logging.ERROR."""
CRITICAL = logging.CRITICAL
"""Constant for logging.CRITICAL."""

# Formatting
STDOUT_MSG_FORMAT = "%(levelname)s - %(instance_name)s -  %(module)s - %(funcName)s - %(message)s"
"""Standard output message format."""

FILE_MSG_FORMAT = STDOUT_MSG_FORMAT
"""File message format."""

DEFAULT_STDOUT_HEADER = """
LEVEL - INSTANCE NAME - MODULE - FUNCTION - MESSAGE
"""
"""Default standard output header."""

DEFAULT_FILE_HEADER = DEFAULT_STDOUT_HEADER
"""Default file header."""

NEW_SESSION_HEADER = f"""
===============================================================================
       NEW SESSION - {datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
==============================================================================="""
"""Default new session header containing date and time."""

string_to_loglevel = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "WARNING": logging.WARN,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


class PyMechanicalCustomAdapter(logging.LoggerAdapter):
    """Keeps the reference to the name of the Mechanical instance dynamic.

    The standard approach supplies extra input to the logger. If this approach
    was used, Mechanical instances would have to be inputted every time a log
    is created.

    Using an adapter means that the reference to the Mechanical instance must only
    be specified once.
    """

    level = (
        None  # This is maintained for compatibility with ``suppress_logging``, but it does nothing.
    )
    file_handler = None
    stdout_handler = None

    def __init__(self, logger, extra=None):
        """Initialize the PyMechanical custom adapter."""
        # super().__init__(logger,extra)

        self.logger = logger
        if extra is not None:
            self.extra = weakref.proxy(extra)
        else:
            self.extra = None
        self.file_handler = logger.file_handler
        self.std_out_handler = logger.std_out_handler

    def process(self, msg, kwargs):
        """Process the message."""
        kwargs["extra"] = {}
        # These are the extra parameters to send to the log.
        kwargs["extra"][
            "instance_name"
        ] = self.extra.name  # Here self.extra is the argument to pass to the log records.
        return msg, kwargs

    def log_to_file(self, filename=FILE_NAME, level=LOG_LEVEL):
        """Add a file handler to the logger.

        Parameters
        ----------
        filename : str, optional
            Name of the file where logs are recorded. The default is ``FILE_NAME``.
        level : str, optional
            Level of logging. The default is ``None``, in which case the ``"DEBUG"``
            level is used. Options are ``"DEBUG"``, ``"INFO"``, ``"WARNING"``,
            and ``"ERROR"``.
        """
        self.logger = addfile_handler(
            self.logger, filename=filename, level=level, write_headers=True
        )
        self.file_handler = self.logger.file_handler

    def log_to_stdout(self, level=LOG_LEVEL):
        """Add a standard output handler to the logger.

        Parameters
        ----------
        level : str, optional
            Level of logging. The default is ``None``, in which case the ``"DEBUG"``
            level is used. Options are ``"DEBUG"``, ``"INFO"``, ``"WARNING"``,
            and ``"ERROR"``.
        """
        if self.std_out_handler:
            raise Exception("Stdout logger is already defined.")

        self.logger = add_stdout_handler(self.logger, level=level)
        self.std_out_handler = self.logger.std_out_handler

    def setLevel(self, level="DEBUG"):
        """Change the log level of the object and the attached handlers.

        Parameters
        ----------
        level : str, optional
            Level of logging. The default is ``"DEBUG"``. Options are ``"DEBUG"``,
            ``"INFO"``, ``"WARNING"``, and ``"ERROR"``.
        """
        if isinstance(level, str):
            level = string_to_loglevel[level.upper()]
        self.logger.setLevel(level)
        for each_handler in self.logger.handlers:
            each_handler.setLevel(level)
        self.level = level


class PyMechanicalPercentStyle(logging.PercentStyle):
    """Controls the way PyMechanical formats the percent style."""

    def __init__(self, fmt, *, defaults=None):
        """Initialize the PyMechanical percent style."""
        # super().__init__(fmt)

        self._fmt = fmt or self.default_format
        self._defaults = defaults

    def _format(self, record):
        defaults = self._defaults
        if defaults:
            values = defaults | record.__dict__
        else:
            values = record.__dict__

        # We can do here any changes we want in record, for example adding a key.

        # We could create an if here if we want conditional formatting, and even
        # change the record.__dict__.
        # Since now we don't want to create conditional fields, it is fine to keep
        # the same MSG_FORMAT for all of them.

        # For the case of logging exceptions to the logger.
        values.setdefault("instance_name", "")

        return STDOUT_MSG_FORMAT % values


class PyMechanicalFormatter(logging.Formatter):
    """Provides for overwriting default format styles with custom format styles.

    Parameters
    ----------
    fmt : optional
        The default is ``STDOUT_MSG_FORMAT``.
    datefmt : optional
        The default is ``None``.
    style : optional
        The default is ``%``.
    validate : bool, optional
        The default is ``None``.
    """

    def __init__(
        self,
        fmt=STDOUT_MSG_FORMAT,
        datefmt=None,
        style="%",
        validate=True,
        defaults=None,
    ):
        """Initialize the PyMechanical formatter."""
        if sys.version_info[1] < 8:  # pragma: no cover
            super().__init__(fmt, datefmt, style)
        else:
            # 3.8: The validate parameter was added
            super().__init__(fmt, datefmt, style, validate)
        self._style = PyMechanicalPercentStyle(fmt, defaults=defaults)  # overwriting


class InstanceFilter(logging.Filter):
    """Ensures that the instance name record always exists."""

    def filter(self, record):
        """Check the log record and return ``True`` to log it or ``False`` to discard it."""
        if not hasattr(record, "instance_name"):
            record.instance_name = ""
        return True


class Logger:
    """Provides for adding handlers to the logger for each Mechanical session.

    This class allows you to add handlers to the logger to output to a file or
    the standard output.

    Parameters
    ----------
    level : int, optional
        Logging level for filtering the messages that are allowed in the logger.
        The default is ``10``, in which case the ``DEBUG`` level is used.
    to_file : bool, optional
        Whether to write log messages to a file. The default is ``False``.
    to_stdout : bool, optional
        Whether to write log messages to the standard output. The default is
        ``True``.
    filename : str, optional
        Name of the file to write log messages to. The default is ``pymechanical.log``.

    Examples
    --------
    Demonstrate logger usage from a Mechanical instance. The logger is automatically
    created when a Mechanical instance is created.

    >>> from ansys.mechanical.core import launch_mechanical
    >>> mechanical = launch_mechanical(loglevel='DEBUG')
    >>> mechanical.log.info('This is a useful message')
    INFO -  -  <ipython-input-24-80df150fe31f> - <module> - This is LOG debug message.

    Import the PyMechanical global logger and add a file output handler.

    >>> import os
    >>> from ansys.mechanical.core import LOG
    >>> file_path = os.path.join(os.getcwd(), 'pymechanical.log')
    >>> LOG.log_to_file(file_path)

    """

    file_handler = None
    std_out_handler = None
    _level = logging.DEBUG
    _instances = {}

    def __init__(self, level=logging.DEBUG, to_file=False, to_stdout=True, filename=FILE_NAME):
        """Customize the logger for PyMechanical.

        Parameters
        ----------
        level : int, optional
            Level of logging that is defined in the ``logging`` package. The default is 'DEBUG'.
            Options are ``"DEBUG"``, ``"INFO"``, ``"WARNING"``, and ``"ERROR"``.
        to_file : bool, optional
            Whether to write log messages to a file. The default is  ``False``.
        to_stdout : bool, optional
            Whether to write log messages to the standard output, which is the
            command line. The default is ``True``.
        filename : str, optional
            Name of the output file. The default is ``pymechanical.log``.
        """
        # Create default main logger.
        self.logger = logging.getLogger("pymechanical_global")
        self.logger.addFilter(InstanceFilter())
        self.logger.setLevel(level)
        self.logger.propagate = True
        self.level = self.logger.level  # TODO: TO REMOVE

        # Write logging methods.
        self.debug = self.logger.debug
        self.info = self.logger.info
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical
        self.log = self.logger.log

        if to_file or filename != FILE_NAME:
            # Record to a file.
            self.log_to_file(filename=filename, level=level)

        if to_stdout:
            self.log_to_stdout(level=level)

        # Use logger to record unhandled exceptions.
        self.add_handling_uncaught_exceptions(self.logger)

    def log_to_file(self, filename=FILE_NAME, level=LOG_LEVEL):
        """Add a file handler to the logger.

        Parameters
        ----------
        filename : str, optional
            Name of the file to write log messages to. The default is
            ``'pymechanical.log'``.
        level : str, optional
            Level of logging. The default is ``10``, in which case the ``"DEBUG"``
            level is used. Options are ``"DEBUG"``, ``"INFO"``,
            ``"WARNING"`` and ``"ERROR"``.

        Examples
        --------
        Write to the ``pymechanical.log`` file in the current working directory.

        >>> from ansys.mechanical.core import LOG
        >>> import os
        >>> file_path = os.path.join(os.getcwd(), 'pymechanical.log')
        >>> LOG.log_to_file(file_path)

        """
        addfile_handler(self, filename=filename, level=level, write_headers=True)

    def log_to_stdout(self, level=LOG_LEVEL):
        """Add a standard output handler to the logger.

        Parameters
        ----------
        level : str, optional
            Level of logging, such as ``DUBUG``. The default is ``LOG_LEVEL``.
        """
        add_stdout_handler(self, level=level)

    def setLevel(self, level="DEBUG"):
        """Change the log level of the object and the attached handlers.

        Parameters
        ----------
        level : str, optional
            Level of logging, such as ``DUBUG``. The default is ``LOG_LEVEL``.
        """
        if isinstance(level, str):
            level = string_to_loglevel[level.upper()]
        self.logger.setLevel(level)
        for each_handler in self.logger.handlers:
            each_handler.setLevel(level)
        self.level = level

    def _make_child_logger(self, suffix, level):
        """Create a child logger.

        Uses the ``getChild`` method to copy attributes between  the ``pymechanical_global``
        logger and a new child logger.

        Parameters
        ----------
        suffix : str
            Name for the child logger.
        level : str, optional
            Level of logging, such as ``DUBUG``. The default is ``LOG_LEVEL``.
        """
        logger = logging.getLogger(suffix)
        logger.std_out_handler = None
        logger.file_handler = None

        if self.logger.hasHandlers:
            for each_handler in self.logger.handlers:
                new_handler = copy(each_handler)

                if each_handler == self.file_handler:
                    logger.file_handler = new_handler
                elif each_handler == self.std_out_handler:
                    logger.std_out_handler = new_handler

                if level:
                    # The logger handlers are copied and changed the loglevel is
                    # the specified log level is lower than the one of the
                    # global.
                    if each_handler.level > string_to_loglevel[level.upper()]:
                        new_handler.setLevel(level)

                logger.addHandler(new_handler)

        if level:
            if isinstance(level, str):
                level = string_to_loglevel[level.upper()]
            logger.setLevel(level)

        else:
            logger.setLevel(self.logger.level)

        logger.propagate = True
        return logger

    def add_child_logger(self, suffix, level=None):
        """Add a child logger to the main logger.

        This child logger is more general than an instance logger, which is designed to
        track the state of a Mechanical instance.

        If the logging level is specified in the arguments, a new logger with a reference
        to the ``_global`` logger handlers is created instead of a child logger.

        Parameters
        ----------
        suffix : str
            Name for the child logger.
        level : str, optional
            Level of logging. The default is ``None``, in which case the ``"DEBUG"``
            level is used. Options are ``"DEBUG"``, ``"INFO"``, ``"WARNING"``,
            and ``"ERROR"``.

        Returns
        -------
        logging.logger
            Logger class.
        """
        name = self.logger.name + "." + suffix
        self._instances[name] = self._make_child_logger(name, level)
        return self._instances[name]

    def _add_mechanical_instance_logger(self, name, mechanical_instance, level):
        if isinstance(name, str):
            instance_logger = PyMechanicalCustomAdapter(
                self._make_child_logger(name, level), mechanical_instance
            )
        elif name is None:
            instance_logger = PyMechanicalCustomAdapter(
                self._make_child_logger("NO_NAMED_YET", level), mechanical_instance
            )
        else:
            raise ValueError("You can only input 'str' classes to this method.")

        return instance_logger

    def add_instance_logger(self, name, mechanical_instance, level=None):
        """Add a logger for a Mechanical instance.

        The logger for a Mechanical instance has an adapter that adds contextual information,
        such as the name of the Mechanical instance. This logger is returned, and you can use it to
        log events as a normal logger. It is also stored in the ``_instances`` field.

        Parameters
        ----------
        name : str
            Name for the new logger.
        mechanical_instance : ansys.mechanical.core.mechanical.Mechanical
            Mechanical instance object. This object should contain the ``name``
            attribute.
        level : str, optional
            Level of logging. The default is ``None``, in which case the ``"DEBUG"``
            level is used. Options are ``"DEBUG"``, ``"INFO"``, ``"WARNING"``,
            and ``"ERROR"``.

        Returns
        -------
        ansys.mechanical.core.logging.PyMechanicalCustomAdapter
            Logger adapter customized to add Mechanical information to the
            logs. You can use this class to log events in the same
            way you would with the ``logger`` class.

        Raises
        ------
        Exception
            You can only input strings as ``name`` to this method.
        """
        count_ = 0
        new_name = name
        while new_name in logging.root.manager.__dict__.keys():
            count_ += 1
            new_name = name + "_" + str(count_)

        self._instances[new_name] = self._add_mechanical_instance_logger(
            new_name, mechanical_instance, level
        )
        return self._instances[new_name]

    def __getitem__(self, key):
        """Get the instance logger based on a key.

        Parameters
        ----------
            key :
        """
        if key in self._instances.keys():
            return self._instances[key]
        else:
            raise KeyError(
                f"There is no instances with name {key}. "
                f"Available keys are {self._instances.keys()}"
            )

    @staticmethod
    def add_handling_uncaught_exceptions(logger):
        """Redirect the output of an exception to a logger.

        Parameters
        ----------
        logger : str
            Name of the logger.
        """

        def handle_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return
            logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

        sys.excepthook = handle_exception


def addfile_handler(logger, filename=FILE_NAME, level=LOG_LEVEL, write_headers=False):
    """Add a file handler to the input.

    Parameters
    ----------
    logger : logging.Logger or logging.Logger
        Logger to add the file handler to.
    filename : str, optional
        Name of the output file. The default is ``FILE_NAME``.
    level : str, optional
        Level of logging. The default is ``None``. Options are ``"DEBUG"``, ``"INFO"``,
        ``"WARNING"`` and ``"ERROR"``.
    write_headers : bool, optional
        Whether to write headers to the file. The default is ``False``.

    Returns
    -------
    logger
        Logger object.
    """
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))

    if isinstance(logger, Logger):
        logger.file_handler = file_handler
        logger.logger.addHandler(file_handler)

    elif isinstance(logger, logging.Logger):
        logger.file_handler = file_handler
        logger.addHandler(file_handler)

    if write_headers:
        file_handler.stream.write(NEW_SESSION_HEADER)
        file_handler.stream.write(DEFAULT_FILE_HEADER)

    return logger


def add_stdout_handler(logger, level=LOG_LEVEL, write_headers=False):
    """Add a file handler to the stand output handler.

    Parameters
    ----------
    logger : logging.Logger or logging.Logger
        Logger to add the file handler to.
    level : str, optional
        Level of logging. The default is ``None``. Options are ``"DEBUG"``, ``"INFO"``,
        ``"WARNING"`` and ``"ERROR"``.
    write_headers : bool, optional
        Whether to write headers to the file. The default is ``False``.

    Returns
    -------
    logger
        Logger object.
    """
    std_out_handler = logging.StreamHandler()
    std_out_handler.setLevel(level)
    std_out_handler.setFormatter(PyMechanicalFormatter(STDOUT_MSG_FORMAT))

    if isinstance(logger, Logger):
        logger.std_out_handler = std_out_handler
        logger.logger.addHandler(std_out_handler)

    elif isinstance(logger, logging.Logger):
        logger.addHandler(std_out_handler)

    if write_headers:
        std_out_handler.stream.write(DEFAULT_STDOUT_HEADER)

    return logger



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""clr_loader for pymechanical embedding. This loads the CLR on both windows and linux."""
import os
import warnings


def __get_mono(assembly_dir, config_dir):
    import clr_loader

    libmono = os.path.join(assembly_dir, "libmonosgen-2.0.so")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        mono = clr_loader.get_mono(
            set_signal_chaining=True,
            libmono=libmono,  # TODO: find_mono is broken on clr-loader v0.2.6
            assembly_dir=assembly_dir,
            config_dir=config_dir,
        )
    return mono


def _set_mono_trace():  # pragma: no cover
    if "ANS_MONO_TRACE" in os.environ:
        os.environ["MONO_LOG_LEVEL"] = "message"
        os.environ["MONO_LOG_MASK"] = "all"


def _load_mono(install_loc):
    """Load the clr using mono that is shipped with the unified install."""
    from pythonnet import load

    _set_mono_trace()
    mono_dir = os.path.join(install_loc, "Tools", "mono", "Linux64")
    assembly_dir = os.path.join(mono_dir, "lib")
    config_dir = os.path.join(mono_dir, "etc")
    mono = __get_mono(assembly_dir, config_dir)
    load(mono)


def load_clr(install_loc: str) -> None:
    """Load the clr, the outcome of this function is that `clr` is usable."""
    if os.name == "nt":  # pragma: no cover
        return
    _load_mono(install_loc)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contain miscellaneous functions and methods at the module level."""

from functools import wraps
import os
import socket
from threading import Thread


def is_windows():
    """Check if the host machine is on Windows.

    Returns
    -------
    ``True`` if the host machine is on Windows, ``False`` otherwise.
    """
    if os.name == "nt":  # pragma: no cover
        return True

    return False


def get_mechanical_bin(release_version):
    """Get the path for the Mechanical executable file based on the release version.

    Parameters
    ----------
    release_version: str
        Mechanical version using the three-digit format. For example, ``"242"`` for
        2024 R2.
    """
    if is_windows():  # pragma: no cover
        program_files = os.getenv("PROGRAMFILES", os.path.join("c:\\", "Program Files"))
        ans_root = os.getenv(
            f"AWP_ROOT{release_version}",
            os.path.join(program_files, "ANSYS Inc", f"v{release_version}"),
        )
        mechanical_bin = os.path.join(ans_root, "aisol", "bin", "winx64", f"AnsysWBU.exe")
    else:
        ans_root = os.getenv(f"AWP_ROOT{release_version}", os.path.join("/", "usr", "ansys_inc"))
        mechanical_bin = os.path.join(*ans_root, f"v{release_version}", "aisol", f".workbench")

    return mechanical_bin


def threaded(func):
    """Decorate a function with this decorator to call it using a thread."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        name = kwargs.get("name", f"Threaded `{func.__name__}` function")
        thread = Thread(target=func, name=name, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper


def threaded_daemon(func):
    """Decorate a function with this decorator to call it using a daemon thread."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        name = kwargs.get("name", f"Threaded (with Daemon) `{func.__name__}` function")
        thread = Thread(target=func, name=name, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()
        return thread

    return wrapper


def check_valid_ip(ip):
    """Check if the IP address is valid.

    Parameters
    ----------
    ip : str
        IP address to check.

    """
    if ip.lower() != "localhost":
        ip = ip.replace('"', "").replace("'", "")
        socket.inet_aton(ip)


def check_valid_port(port, lower_bound=1000, high_bound=60000):
    """Check if the port is valid.

    Parameters
    ----------
    port : int
        Port to check.
    lower_bound : int, optional
       Lowest possible value for the port. The default is ``1000``.
    high_bound : int, optional
       Highest possible value for the port. The default is ``60000``.
    """
    if not isinstance(port, int):
        raise ValueError("The 'port' parameter must be an integer.")

    if lower_bound < port < high_bound:
        return
    else:
        raise ValueError(f"'port' values must be between {lower_bound} and {high_bound}.")


def check_valid_start_instance(start_instance):
    """
    Check if the value obtained from the environmental variable is valid.

    Parameters
    ----------
    start_instance : str, bool
        Value obtained from the corresponding environment variable.

    Returns
    -------
    bool
        ``True`` if ``start_instance`` is ``True`` or ``"True"``,
        ``False`` otherwise.

    """
    if not isinstance(start_instance, (str, bool)):
        raise ValueError("The value for 'start_instance' should be a string or a boolean.")

    if isinstance(start_instance, bool):
        return start_instance

    if start_instance.lower() not in ["true", "false"]:
        raise ValueError(
            f"The value for 'start_instance' should be 'True' or 'False' (case insensitive)."
        )

    return start_instance.lower() == "true"


def is_float(input_string):
    """Check if a string can be converted to a float.

    Parameters
    ----------
    input_string : str
        String to check.

    Returns
    -------
    bool
        ``True`` when conversion is possible, ``False`` otherwise.
    """
    try:
        float(input_string)
        return True
    except ValueError:
        return False
